package core

import (
	"validator/quest"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	validatorv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/validator/v1"
	log "github.com/sirupsen/logrus"
)

// LoadPendingProofs 从questdb读取当天proofs并验证
func (v *Validator) LoadPendingProofs(startTimestamp, endTimestamp int64) {
	defaultLimit := 100
	lastTaskID := ""
	var dbProofs []*quest.ProofModel
	var err error
	for {
		dbProofs, err = v.q.GetProofs(startTimestamp, endTimestamp, lastTaskID, defaultLimit)
		if err != nil {
			log.WithError(err).Error("failed to get proofs")
			return
		}
		log.WithField("proof count", len(dbProofs)).Debug("load pending proofs")

		if len(dbProofs) == 0 {
			return
		}

		for _, dbProof := range dbProofs {
			miner, proof := v.verifyProof(dbProof)
			if proof != nil {
				v.AddPendingProof(miner, proof)
			}
			lastTaskID = dbProof.TaskId
		}
	}
}

func (v *Validator) LoadMockProofs() {
	mockAddresses := []common.Address{
		common.HexToAddress("0x1111111111111111111111111111111111111111"),
		common.HexToAddress("0x2222222222222222222222222222222222222222"),
		common.HexToAddress("0x3333333333333333333333333333333333333333"),
		common.HexToAddress("0x4444444444444444444444444444444444444444"),
		common.HexToAddress("0x5555555555555555555555555555555555555555"),
		common.HexToAddress("0x6666666666666666666666666666666666666666"),
		common.HexToAddress("0x7777777777777777777777777777777777777777"),
		common.HexToAddress("0x8888888888888888888888888888888888888888"),
		common.HexToAddress("0x9999999999999999999999999999999999999999"),
		common.HexToAddress("0x99999999953AF2bC757C30824288A706D45fB44b"),
	}

	proof := &validatorv1.ValidatedProof{
		Workload: 1000,
	}

	for _, address := range mockAddresses {
		v.AddPendingProof(address, proof)
	}
}

func (v *Validator) verifyProof(dbProof *quest.ProofModel) (miner common.Address, proof *validatorv1.ValidatedProof) {
	if dbProof.TaskWorkload == 0 {
		return
	}
	containerPubKey, err := crypto.SigToPub(dbProof.GenerateHashPayload(), common.Hex2Bytes(dbProof.TaskContainerSignature))
	if err != nil {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId}).WithError(err).Error("failed to verify container signature")
		return
	}
	containerSigner := crypto.PubkeyToAddress(*containerPubKey)
	verified := v.IsContainerAddress(containerSigner)
	if !verified {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId, "signer": containerSigner.Hex()}).Error("invalid container signature")
		return
	}

	minerPubKey, err := crypto.SigToPub(dbProof.GenerateHashPayload(), common.Hex2Bytes(dbProof.TaskMinerSignature))
	if err != nil {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId}).WithError(err).Error("failed to verify miner signature")
		return
	}
	minerAddress := crypto.PubkeyToAddress(*minerPubKey)
	if minerAddress.Hex() != common.HexToAddress(dbProof.TaskWorkerAccount).Hex() {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId, "miner": dbProof.TaskWorkerAccount, "recovered miner": minerAddress.Hex()}).Error("invalid miner address")
		return
	}

	nmPubKey, err := crypto.SigToPub(dbProof.GenerateNMHashPayload(), common.Hex2Bytes(dbProof.TaskManagerSignature))
	if err != nil {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId}).WithError(err).Error("failed to verify manager signature")
		return
	}

	nmAddress := crypto.PubkeyToAddress(*nmPubKey)
	verified = v.IsNMAddress(nmAddress)
	if !verified {
		log.WithFields(log.Fields{"taskid": dbProof.TaskId, "signer": nmAddress.Hex()}).Error("invalid manager signature")
		return
	}

	proof = &validatorv1.ValidatedProof{
		Workload:  dbProof.TaskWorkload,
		Timestamp: uint64(dbProof.TaskFinishTimestamp),
	}
	return common.HexToAddress(dbProof.TaskProfitAccount), proof
}

func (v *Validator) GetGlobalWorkload() uint64 {
	workload, err := v.q.GetGlobalWorkload(v.todayTimestamp())
	if err != nil {
		log.WithError(err).Error("failed to get global workload")
		return 0
	}
	log.WithField("workload", workload).Debug("quest get global workload")
	return workload
}
