package core

import (
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/core/rawdb"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/ethdb"
	"github.com/ethereum/go-ethereum/trie"
	"github.com/ethereum/go-ethereum/trie/trienode"
	"github.com/gogo/protobuf/proto"
	validatorv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/validator/v1"
	log "github.com/sirupsen/logrus"
)

type StateDB struct {
	db   *trie.Database
	trie *trie.Trie
}

func NewStateDB(lvdb ethdb.KeyValueStore, root common.Hash) (statedb *StateDB, err error) {
	trieDB := trie.NewDatabase(rawdb.NewDatabase(lvdb), nil)
	// owner 如果是0x0，启用的是用户树，此处并没有使用 stateaccount
	t, err := trie.New(trieID(root), trieDB)
	if err != nil {
		return nil, err
	}
	return &StateDB{
		db:   trieDB,
		trie: t,
	}, nil
}

func (s *StateDB) GetMinerObject(miner common.Address) (object *validatorv1.MinerObject) {
	object = &validatorv1.MinerObject{
		Miner:   miner.Hex(),
		Balance: "0",
	}
	k := crypto.Keccak256(miner.Bytes())
	v, err := s.trie.Get(k)
	if err != nil {
		log.WithError(err).Warn("trie get miner object failed")
		return object
	}
	log.Debugf("GetMinerObject %s %s\n", miner.Hex(), common.Bytes2Hex(v))
	if v == nil {
		return object
	}
	err = proto.Unmarshal(v, object)
	if err != nil {
		return nil
	}
	return
}

func (s *StateDB) UpdateMinerObject(miner common.Address, object *validatorv1.MinerObject) (err error) {
	k := crypto.Keccak256(miner.Bytes())
	object.Miner = miner.Hex()
	v, err := proto.Marshal(object)
	if err != nil {
		return
	}

	err = s.trie.Update(k, v)
	return
}

func (s *StateDB) Commit() (root common.Hash, err error) {
	root, nodes, err := s.trie.Commit(false)
	if err != nil {
		return
	}
	nodesSet := trienode.NewMergedNodeSet()
	if nodes != nil {
		err = nodesSet.Merge(nodes)
		if err != nil {
			return
		}
	}
	// block states 在 hashtrie 中未使用
	err = s.db.Update(root, common.Hash{}, 0, nodesSet, nil)
	if err != nil {
		return
	}

	err = s.db.Commit(root, true)
	return root, err
}

func (s *StateDB) IterAllObject() (objects []*validatorv1.MinerObject) {
	iter, err := s.trie.NodeIterator(nil)
	if err != nil {
		return
	}
	for iter.Next(true) {
		if !iter.Leaf() {
			continue
		}
		v := iter.LeafBlob()
		object := &validatorv1.MinerObject{}
		err := proto.Unmarshal(v, object)
		if err != nil {
			log.WithError(err).Error("failed to unmarshal object, iter")
			continue
		}
		log.WithFields(log.Fields{
			"miner":    object.Miner,
			"balance":  object.Balance,
			"workload": object.Workload,
		}).Debug("iter miner object")
		objects = append(objects, object)
	}
	return
}

func trieID(root common.Hash) (id *trie.ID) {
	return &trie.ID{
		StateRoot: root,
		Owner:     common.HexToHash("0x1"),
		Root:      root,
	}
}
